//==========================================================================
// File: face.c
// Description: An implementation of Singh and Urrutia routing.
// Author: Pat Morin
// Created: Sun Apr  4 07:26:37 EDT 1999 (Yes, easter morning)
//==========================================================================
#include "face.h"

#include <LEDA/segment.h>
#include <LEDA/ray.h>
#include <LEDA/face_map.h>

// #define WIN

#include <LEDA/window.h>

// Pi * 2 (from Knuth)
#define PI_TIMES_2 (3.1415926535897932384626433832795028841972*2.0)

// #define WIN

#ifdef WIN
static window W(600+20, 600+20, "FACE_PATH debug window");  
#endif // WIN

//==========================================================================
// Find the first edge in the FACE-1 or FACE-2 path from s to t
//==========================================================================
static edge find_first(graph &G, node_array<point> &pos,  node s, node t)
{
  edge e, emax = nil;
  double alphamax = -HUGE_VAL;
  forall_adj_edges (e, s) {
    double a1 = ray(pos[s], pos[t]).direction();
    double a2 = ray(pos[s], pos[G.target(e)]).direction();
    double alpha;
    if (a2 > a1) {
      alpha = a2 - a1;
    } else {
      alpha = a2 + PI_TIMES_2 - a1;
    }
    if (alpha == 0.0) alpha = PI_TIMES_2;
    if (alpha > alphamax) {
      alphamax = alpha;
      emax = e;
    }
  }
  return emax;
}

//==========================================================================
// Find the KSU path from s to t
// Complexity: O(n)
// Precondition: G is an embedded planar map
//========================================================================
void KSU_PATH(graph &G, node_array<point> &pos, 
	      node s, node t, list<edge> &P)
{
  edge e;
  window W(600+20, 600+20, "unit graph debug window");  
  W.button("dismiss",1);
  W.open(window::center,window::center);
  W.init(-.1, 1.2, -.1);
  W.set_node_width(3);
  node v;
  forall_nodes (v, G) {
    W.draw_filled_node(pos[v], blue);
  }
  forall_edges (e, G) {
    W.draw_segment(pos[G.source(e)], pos[G.target(e)], black);
  }
  W.draw_filled_node(pos[s], red);
  W.draw_filled_node(pos[t], green);
  W.set_line_width(3);
  W.set_mode(xor_mode);

  P.clear();

  // first check if s==t or s is adjacent to t
  if (s == t) return;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects segment (s,t)
  e = find_first(G, pos, s, t);

  segment st(pos[s], pos[t]);
  // start walking
  point p = pos[s];
  while (1 < 2) {
    // Find the entry edge of the current face
    segment pt(p, pos[t]);
    edge start = e, entry = nil;
    W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
    int count = 0, entry_count = 0;
    do {
      P.append(e);
      if (G.target(e) == t) return;
      edge eold = e;
      e = G.face_cycle_succ(e);
      W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
      W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
      count++;
      point px;
      segment se(pos[G.source(e)], pos[G.target(e)]);
      if (collinear(p, pos[G.target(e)], pos[t])
	  && pos[G.target(e)].distance(pos[t]) < p.distance(pos[t])) {
	entry_count = count;
	point pold = p;
	p = pos[G.target(e)];
	W.draw_filled_node(p, orange);
	W.draw_filled_node(pold, orange);
	entry = e;
      } else if (st.intersection(pt, px)
		 && px.distance(pos[t]) < p.distance(pos[t])) {
	entry_count = count;
	point pold = p;
	p = px;
	W.draw_filled_node(p, orange);
	W.draw_filled_node(pold, orange);
	entry = e;
      }
    } while (e != start);

    // return to the entry edge
    while (e != entry) {
      if (entry_count > count/2) {
	P.append(G.reversal(e));
	edge eold = e;
	e = G.face_cycle_pred(e);
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
      } else {
	P.append(e);
	edge eold = e;
	e = G.face_cycle_succ(e);
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
      }
    }
    e = G.cyclic_adj_pred(e);
  }
}


//==========================================================================
// Find the BKOO path from s to t
// Complexity: O(n)
// Precondition: G is an embedded planar map
//========================================================================
void FACE_TREE_PATH(graph &G, node_array<point> &pos, 
		    node s, node t, list<edge> &P)
{
  enum {EDGE, VERTEX};

  P.clear();
  
  G.compute_faces();
  face_map<bool> visited(G);
  face f;
  forall_faces (f, G) {
    visited[f] = false;
  }

  // first check if s==t or s is adjacent to t
  if (s == t) return;
  edge e;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects segment (s,t)
  e = find_first(G, pos, s, t);

  // start walking
  while (1 < 2) {
    // Find the entry edge of the current face
    edge start = e, eentry = nil;
    if (visited[G.face_of(e)]) {
      cerr << "FACE_TREE_PATH: Visiting the same face twice!" << endl;
      edge e;
      window W(600+20, 600+20, "unit graph debug window");  
      W.button("dismiss",1);
      W.open(window::center,window::center);
      W.init(-.1, 1.2, -.1);
      W.set_node_width(3);
      node v;
      forall_nodes (v, G) {
	W.draw_filled_node(pos[v], blue);
      }
      forall_edges (e, G) {
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], black);
      }
      W.draw_filled_node(pos[s], red);
      W.draw_filled_node(pos[t], green);
      W.set_line_width(3);
      W.set_mode(xor_mode);

      W.draw_segment(pos[G.source(P.front())], 
		     pos[G.target(P.front())], orange);
      for (list_item it = P.succ(P.first()); it != nil; it = P.succ(it)) {
	W.draw_segment(pos[G.source(P[it])], 
		       pos[G.target(P[it])], orange);
	W.draw_segment(pos[G.source(P[P.pred(it)])], 
		       pos[G.target(P[P.pred(it)])], orange);
      }
    }
    visited[G.face_of(e)] = true;
    node ventry = nil;
    int count = 0, entry_count = 0;
    double dmin = HUGE_VAL;
    int type;
    do {
      P.append(e);
      if (G.target(e) == t) return;
      e = G.face_cycle_succ(e);
      node v;
      count++;
      double d;
      point p1 = pos[G.source(e)];
      point p2 = pos[G.target(e)];
      point px;
      int tmptype;
      line perp(line(p1, p2).perpendicular(pos[t]));
      if (perp.intersection(segment(p1, p2), px) 
	  && px.distance(p1) > 0.00001
	  && px.distance(p2) > 0.00001) {
	d = pos[t].distance(px);
	tmptype = EDGE;
      } else {
	tmptype = VERTEX;
	double d1 = p1.distance(pos[t]);
	double d2 = p2.distance(pos[t]);
	if (d1 < d2) {
	  d = d1;
	  v = G.source(e);
	} else {
	  d = d2;
	  v = G.target(e);
	}
      }
      if ((dmin - d > .0001) ||
	  (fabs(d - dmin) < 0.00001 && tmptype == VERTEX)) {
	type = tmptype;
	eentry = e;
	ventry = v;
	entry_count = count;
	dmin = d;
      }
    } while (e != start);

    // return to the entry edge/vertex
    if (type == EDGE) {
      if (entry_count > count/2) {
	e = G.face_cycle_pred(e);
	while (e != eentry) {
	  P.append(G.reversal(e));
	  e = G.face_cycle_pred(e);
	}
	e = G.reversal(e);
      } else {
	while (e != eentry) {
	  P.append(e);
	  e = G.face_cycle_succ(e);
	}
	e = G.cyclic_adj_pred(e);
      }
    } else { // type == VERTEX
      if (entry_count > count/2) {
	e = G.face_cycle_succ(e);
	while (G.target(e) != ventry) {
	  P.append(G.reversal(e));
	  e = G.face_cycle_pred(e);
	}
      } else {
	while (G.source(e) != ventry) {
	  P.append(e);
	  e = G.face_cycle_succ(e);
	}	
      }
      e = find_first(G, pos, ventry, t);
    }
  }
}



//==========================================================================
// Find the BKOO path from s to t
// Complexity: O(n)
// Precondition: G is an embedded planar map
//========================================================================
void BKOO_PATH(graph &G, node_array<point> &pos, 
	       node s, node t, list<edge> &P)
{
  edge e;

  P.clear();

  G.compute_faces();
  face_map<bool> visited(G);
  face f;
  forall_faces (f, G) {
    visited[f] = false;
  }
  
  // first check if s==t or s is adjacent to t
  if (s == t) return;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects segment (s,t)
  e = find_first(G, pos, s, t);

  // start walking
  while (1 < 2) {
    // Find the entry edge of the current face
    edge start = e, entry = nil;
    visited[G.face_of(start)] = true;
    int count = 0, entry_count = 0;
    double dmin = HUGE_VAL;
    bool just_set = false;
    do {
      edge etmp = e;
      P.append(e);
      if (G.target(e) == t) return;
      e = G.face_cycle_succ(e);
      count++;
      double d;
      point p1 = pos[G.source(e)];
      point p2 = pos[G.target(e)];
      point px;
      line perp(line(p1, p2).perpendicular(pos[t]));
      if (perp.intersection(segment(p1, p2), px)) {
	d = pos[t].distance(px);
      } else {
	double d1 = p1.distance(pos[t]);
	double d2 = p2.distance(pos[t]);
	d = (d1 < d2) ? d1 : d2;
      }
      if (((d < dmin) 
	  || (d == dmin && just_set))
	  && !left_turn(p1, p2, pos[t])) {
	just_set = true;
	entry_count = count;
	dmin = d;
	entry = e;
      } else {
	just_set = false;
      }
    } while (e != start);

    // return to the entry edge
    while (e != entry) {
      if (entry_count > count/2) {
	e = G.face_cycle_pred(e);
	P.append(G.reversal(e));
      } else {
	e = G.face_cycle_succ(e);
	P.append(e);
      }
    }
    e = G.cyclic_adj_pred(e);
    if (visited[G.face_of(e)]) {
      cerr << "BKOO_PATH: Visiting the same face twice!" << endl;
      edge e;
      window W(600+20, 600+20, "unit graph debug window");  
      W.button("dismiss",1);
      W.open(window::center,window::center);
      W.init(-.1, 1.2, -.1);
      W.set_node_width(3);
      node v;
      forall_nodes (v, G) {
	W.draw_filled_node(pos[v], blue);
      }
      forall_edges (e, G) {
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], black);
      }
      W.draw_filled_node(pos[s], red);
      W.draw_filled_node(pos[t], green);
      W.set_line_width(3);
      W.set_mode(xor_mode);

      W.draw_segment(pos[G.source(P.front())], 
		     pos[G.target(P.front())], orange);
      for (list_item it = P.succ(P.first()); it != nil; it = P.succ(it)) {
	W.draw_segment(pos[G.source(P[it])], 
		       pos[G.target(P[it])], orange);
	W.draw_segment(pos[G.source(P[P.pred(it)])], 
		       pos[G.target(P[P.pred(it)])], orange);
      }
    }
  }
}

//==========================================================================
// Find the Modified KSU path from s to t
// Complexity: O(n^2)
// Precondition: G is an embedded planar map
//========================================================================
void MKSU_PATH(graph &G, node_array<point> &pos, 
		node s, node t, list<edge> &P)
{
  P.clear();

  if (s == t) return;

  // first check if s is adjacent to t
  edge e;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects segment (s,t)
  e = find_first(G, pos, s, t);

  // start walking
  point p = pos[s];
  segment pt(p, pos[t]);
  edge estart = e;
  while (1 < 2) {
    P.append(e);
    if (G.target(e) == t) return;
    e = G.face_cycle_succ(e);
    point pp;
    while (e != estart && e != G.reversal(estart) && G.target(e) != t 
	   && G.target(e) != s && G.source(e) != s
	   && segment(pos[G.source(e)], pos[G.target(e)]).intersection(pt, pp)
	   && right_turn(pos[G.source(e)], pos[G.target(e)], pos[t])) {
      estart = e;
      e = G.cyclic_adj_pred(e);
      p = pp;
      pt = segment(p, pos[t]);
    }
  }
}





