//==========================================================================
// File: mksu.c
// Description: An implementation of modified Kranakis, Singh and 
//              Urrutia routing
// Author: Pat Morin
// Created: 
//==========================================================================
#include <LEDA/segment.h>

#include "mksu.h"
#include "util.h"

// Use this define to get a debug window that pops up whenever this
// function is called.
// #define DBGWIN

#ifdef DBGWIN
#include <LEDA/window.h>
#endif DBGWIN

//==========================================================================
// Find the MKSU path from s to t
// Complexity: O(n)
//========================================================================
void MKSU_PATH(const graph &G, const node_array<POINT> &pos, 
	      node s, node t, list<edge> &P)
{
  edge e;
#ifdef DBGWIN
  window W(300+10, 300+20, "unit graph debug window");  
  W.button("dismiss",1);
  W.open(window::center,window::center);
  W.init(-.1, 1.2, -.1);
  W.set_node_width(3);
  node v;
  forall_nodes (v, G) {
    W.draw_filled_node(pos[v], blue);
  }
  forall_edges (e, G) {
    W.draw_segment(pos[G.source(e)], pos[G.target(e)], black);
  }
  W.draw_filled_node(pos[s], red);
  W.draw_filled_node(pos[t], green);
  W.set_line_width(3);
  W.set_mode(xor_mode);
#endif // DBGWIN

  P.clear();

  // first check if s==t or s is adjacent to t
  if (s == t) return;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects SEGMENT (s,t)
  e = find_first(G, pos, s, t);

  // always refer to the original segment (s,t) to avoid cumulative
  // arithmetic errors
  SEGMENT st(pos[s], pos[t]);

  // start walking
  POINT p = pos[s];
  while (1 < 2) {
    // Find the entry edge of the current face
    edge eold = nil, start = e;
    while (2 < 3) {
#ifdef DBGWIN
      W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
      if (eold != nil)
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
      P.append(e);
      if (G.target(e) == t) {
	if (!valid_path(G, P, s, t)) {
	  cerr << "MKSU_PATH: returning an invalid path" << endl;
	  forall (e, P) {
	    cerr << pos[G.source(e)] << " " << pos[G.target(e)] << endl;
	  }
	  P.clear();
	}
	return;
      }
      POINT px;
      SEGMENT se(pos[G.source(e)], pos[G.target(e)]);
      int type = robust_intersection_test(st, se, px);
      if (type == I_NONE || type == I_SOURCE) {
	px = POINT(HUGE_VAL, HUGE_VAL);
      }
      if (type == I_OVERLAP) {
	type = I_TARGET;
	px = pos[G.target(e)];
      }
      if (type == I_TARGET && G.target(e) == G.source(start)
	  && G.face_cycle_succ(e) != start) {
	// never pass through the start vertex until we've been 
	// around the whole face
	px = POINT(HUGE_VAL, HUGE_VAL);
      }
      if (p.sqr_dist(pos[t]) - px.sqr_dist(pos[t]) > EPSILON_SQUARED
	  || (px.sqr_dist(pos[t]) == p.sqr_dist(pos[t])
	      && type == I_TARGET)) {
	if (type == I_TARGET) {
	  p = px;
	  eold = e;
	  e = find_first(G, pos, G.target(e), t);
	} else {
	  P.del(P.last());
	  p = px;
	  eold = e;
	  e = G.cyclic_adj_pred(e);
	}
	break;
      }
      eold = e;
      e = G.face_cycle_succ(e);
    }
  }
}



