//==========================================================================
// File: tests.c
// Description: Tests routing strategies for unit graphs
// Author: Pat Morin
// Created: Mon Mar 29 19:09:25 EST 1999
//==========================================================================
#include <iostream.h>
#include <ctype.h>
#include <math.h>

#include <LEDA/list.h>
#include <LEDA/point.h>
#include <LEDA/circle.h>
#include <LEDA/random_source.h>
#include <LEDA/graph.h>
#include <LEDA/graph_alg.h>
#include <LEDA/graph_misc.h>
#include <LEDA/graph_gen.h>
#include <LEDA/plane_alg.h>

#define WIN

#ifdef WIN
#include <LEDA/window.h>
#endif // WIN

#include "mesh.h"
#include "ksu.h"
#include "mksu.h"
#include "facetree.h"


/* 
 * These are the algorithms we test:
 * KSU - The Kranakis, Singh, Urrutia algorithm
 * BKOO - An algorithm based on the BKOO subdivision traversal algorithm
 * FACETREE - An algorithm based on the face tree
 * MKSU - The modified KSU algorithm
 */
#define NALGS 4
enum algorithm { A_KSU, A_BKOO, A_FACETREE, A_MKSU };


/*
 * Various types of test graphs
 */
enum graphtype { G_DELAUNAY, G_GRAHAM, G_MESH };


/*
 * These are colors we use to label connected components in our display
 */
#define NCOLS 16
color colors[NCOLS] =
{ black, white, red, green, blue, yellow, violet, orange, cyan, 
  brown, pink, green2, blue2, grey1, grey2, grey3 };


/* 
 * Run the tests and ouput the results in tabular form.  The results
 * for each graph are output on a single line in the form:
 * <nnodes> <avg-dilation0> <avg-dilation1> ...
 * where the numbers correspond to the order in the enumeration type 
 * algorithm.  
 * The usage is: tests <n> <avgdeg> <nreps>.
 */
int main(int argc, char *argv[]) 
{
  node v;
  edge e;

  // read command line arguments
  if (argc != 6) {
    cerr << "Usage: " << argv[0] << " <graph-type> <n> <vfaults>"
	 << "<efaults> <nreps>" << endl
	 << " Where <graph-type> is one of delaunay, graham or mesh." 
	 << endl;
    exit(1);
  }
  int n = atoi(argv[2]);
  graphtype gtype;
  switch (toupper(argv[1][0])) {
  case 'D':
    gtype = G_DELAUNAY;
    break;
  case 'G':
    gtype = G_GRAHAM;
    break;
  case 'M':
    n = (int) ceil(sqrt(n));
    n *= n;
    gtype = G_MESH;
    break;
  default:
    cerr << "Invalid graph type" << endl;
    exit(1);
    break;
  }
  int vfaults = atoi(argv[3]);
  int efaults = atoi(argv[4]);
  int nreps = atoi(argv[5]);

  // create our persistent data structures
  random_source R;
  GRAPH<point,int> G;
  list<point> L;

#ifdef WIN
    // open the display window
    window W(300+10, 300+20, "unit graph debug window");  
    W.button("dismiss",1);
    W.open(window::center,window::center);
    W.init(-.1, 1.2, -.1);
#endif // WIN

  // we do nreps repetitions of the same experiment
  for (int round = 0; round < nreps; round++) {

    // generate a random test graph
    L.clear();
    double x, y;
    for (int i = 0; i < n; i++) {
      R >> x;
      R >> y;
      L.append(point(x, y));
    }
    switch(gtype) {
    case G_DELAUNAY:
      DELAUNAY_TRIANG(L, G);
      break;
    case G_GRAHAM:
      TRIANGULATE_POINTS(L, G);
      break;
    case G_MESH:
      MESH(n, G);
    }
    // add vertex faults
    for (int i = 0; i < vfaults; i++) {
      v = G.choose_node();
      G.del_node(v);
    }
    // add edge faults
    for (int i = 0; i < efaults; i++) {
      e = G.choose_edge();
      G.del_edge(G.reversal(e));
      G.del_edge(e);
    }
    // compute the connected components
    node_array<int> compnum(G);
    COMPONENTS(G, compnum);

#ifdef WIN
    // display the graph
    W.clear();
    forall_edges (e, G) {
      W.draw_segment(G[G.source(e)], G[G.target(e)], black);
    }
    W.set_node_width(2);
    forall_nodes (v, G) {
      W.draw_filled_node(G[v], colors[compnum[v]%16]);
    }
#ifdef PAUSE
    // wait for the user to continue
    while (1 < 2) { 
      int k = W.read_mouse();
      if (k==1) break;
    }
#endif // PAUSE
#endif // WIN

    // setup the cost, dist and pred arrays for SHORTEST_PATH
    edge_array<double> c(G);
    forall_edges (e, G) {
      c[e] = 1.0;
    }
    node_array<double> dist(G);
    node_array<edge> pred(G);

    // setup the position array for our path finding algorithms
    node v;
    node_array<point> pos(G);
    forall_nodes (v, G) {
      pos[v] = G[v];
    }

    // Keep the sum for the average dilation
    double sum[NALGS];
    int successes[NALGS];
    for (int i = 0; i < NALGS; i++) {
      sum[i] = 0.0;
      successes[i] = 0;
    }

    // Test all pairs of nodes
    node s, t;  // source and target
    forall_nodes (t, G) {
      SHORTEST_PATH(G, t, c, dist, pred);
      forall_nodes (s, G) {
	if (s == t || compnum[s] != compnum[t] || dist[s] < 2) continue;

	list<edge> P;
	KSU_PATH(G, pos, s, t, P);
	if (!P.empty()) {
	  successes[A_KSU]++;
	  sum[A_KSU] += P.length() / dist[s];
	}

	P.clear();
	// BKOO_PATH(G, pos, s, t, P);
	sum[A_BKOO] += P.length() / dist[s];

	P.clear();
	FACETREE_PATH(G, pos, s, t, P);
	if (!P.empty()) {
	  successes[A_FACETREE]++;
	  sum[A_FACETREE] += P.length() / dist[s];
	}

	P.clear();
	// MKSU_PATH(G, pos, s, t, P);
	if (!P.empty()) {
	  successes[A_MKSU]++;
	  sum[A_MKSU] += P.length() / dist[s];
	}
      }
    }

    // Now we have all our results for this run, output them
    cout << n;
    for (int i = 0; i < NALGS; i++) {
      cout << "\t" << sum[i]/successes[i];
    }
    cout << endl;
  } 
  return 0;
}

